from typing import Any, Callable, Iterable, Optional

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import NDArray

from .typing import FilePath

__all__ = ["execute_parallel", "reduce_parallel"]


def execute_parallel(
    function: Callable[[FilePath, FilePath], None],
    srcs: Iterable[FilePath],
    dsts: Iterable[FilePath],
    *,
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> None:
    """Execute function in parallel."""

    delayed_function = delayed(function)
    with Parallel(n_jobs=n_jobs, **kwargs) as parallel:
        parallel(delayed_function(src, dst) for src, dst in zip(srcs, dsts))


def reduce_parallel(
    function: Callable[[FilePath], NDArray],
    srcs: Iterable[FilePath],
    *,
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> NDArray:
    """Reduce function in parallel."""

    delayed_function = delayed(function)
    with Parallel(n_jobs=n_jobs, **kwargs) as parallel:
        result = parallel(delayed_function(src) for src in srcs)
    return np.stack(result)



from typing import Tuple

import numpy as np
import torch
from torch import Tensor

from ...typing import FilePath
from ._loader import Loader

__all__ = ["BinaryLoader"]


class BinaryLoader(Loader):
    def __init__(self, length: int, padding_value: int = 256) -> None:
        self.length = length
        self.padding_value = padding_value

    def __call__(self, path: FilePath) -> Tensor:
        with open(path, "rb") as f:
            buffer = f.read(self.length)
        return self._from_buffer(buffer)

    def from_buffer(self, buffer: bytes) -> Tensor:
        return self._from_buffer(buffer[: self.length])

    def _from_buffer(self, buffer: bytes) -> Tensor:
        # Since PyTorch embedding layer requires int32/int64 as input, we have
        # to convert uint8 to int32.
        binary = np.frombuffer(buffer, dtype=np.uint8)
        binary = binary.astype(np.int32)

        padding_length = self.length - len(binary)
        if padding_length > 0:
            padding = np.full(padding_length, self.padding_value, dtype=np.int32)
            binary = np.concatenate([binary, padding])
        return torch.tensor(binary)

    def _get_args(self) -> Tuple[str, ...]:
        return ("length", "padding_value")



from typing import Tuple

from torch import Tensor

from ...typing import FilePath


class Loader:
    def __call__(self, path: FilePath) -> Tensor:
        raise NotImplementedError

    def __repr__(self) -> str:
        # Generate pretty formatted arguments.
        args = self._get_args()
        args_str = ", ".join(f"{a}={getattr(self, a)!r}" for a in args)
        return f"{type(self).__name__}({args_str})"

    def _get_args(self) -> Tuple[str, ...]:
        raise NotImplementedError



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import torch
from PIL import Image
from torch.utils.data import Dataset

from ..typing import FilePath

__all__ = [
    "LabeledDataset",
    "LabeledByteSeqDataset",
    "LabeledImageDataset",
    "UnlabeledDataset",
    "UnlabeledByteSeqDataset",
    "UnlabeledImageDataset",
]


class ByteSeqLoader:
    def __init__(self, length: int, padding: int = 256) -> None:
        self.length = length
        self.padding = padding

    def __call__(self, path: FilePath) -> torch.Tensor:
        with open(path, "rb") as f:
            buffer = f.read(self.length)

        # Since PyTorch embedding layer requires int32/int64 as input, we have
        # to convert uint8 to int32.
        byte_seq: np.ndarray = np.frombuffer(buffer, dtype=np.uint8)
        byte_seq = byte_seq.astype(np.int32)

        padding_length = self.length - len(byte_seq)
        if padding_length > 0:
            padding_seq = np.full(padding_length, self.padding, dtype=np.int32)
            byte_seq = np.concatenate([byte_seq, padding_seq])
        return torch.tensor(byte_seq)


class PILLoader:
    def __init__(self, mode: Optional[str] = None) -> None:
        self.mode = mode

    def __call__(self, path: FilePath) -> Image.Image:
        with open(path, "rb") as f:
            image = Image.open(f)
            return image.convert(self.mode)


class LabeledDataset(Dataset):
    def __init__(
        self,
        root: FilePath,
        loader: Callable[[FilePath], Any],
        labels: pd.DataFrame,
        *,
        cat: bool = True,
        suffix: Optional[str] = None,
    ) -> None:
        self.root = Path(root)
        self.loader = loader
        self.labels = labels
        self.cat = cat
        self.suffix = suffix

        self._sample_names = [str(item) for item in labels.iloc[:, 0]]
        self._target_names = [str(item) for item in labels.iloc[:, 1]]

        self._index_to_class = sorted(set(self._target_names))
        self._class_to_index = {c: i for i, c in enumerate(self._index_to_class)}

    @property
    def sample_names(self) -> List[str]:
        return self._sample_names

    @property
    def target_names(self) -> List[str]:
        return self._target_names

    @property
    def index_to_class(self) -> List[str]:
        return self._index_to_class

    @property
    def class_to_index(self) -> Dict[str, int]:
        return self._class_to_index

    def __getitem__(self, index: int) -> Tuple[Any, int]:
        sample_name = self._sample_names[index]
        target_name = self._target_names[index]

        if self.cat:
            sample_path = self.root / target_name / sample_name
        else:
            sample_path = self.root / sample_name

        # If `suffix` is an empty string, the suffix will be removed.
        if self.suffix is not None:
            sample_path = sample_path.with_suffix(self.suffix)

        sample = self.loader(sample_path)
        target = self._class_to_index[target_name]
        return sample, target

    def __len__(self) -> int:
        return len(self.labels)

    def __repr__(self) -> str:
        args = ", ".join(f"{a}={getattr(self, a)!r}" for a in self._args())
        s = f"{type(self).__name__}({args}):\n"
        s += f"    Root directory: {self.root}\n"
        s += f"    Number of samples: {len(self)}\n"
        s += f"    Number of classes: {len(self.index_to_class)}\n"
        return s

    def _args(self) -> Tuple[str, ...]:
        return ("cat", "suffix")


class LabeledByteSeqDataset(LabeledDataset):
    def __init__(
        self,
        root: FilePath,
        labels: pd.DataFrame,
        *,
        cat: bool = True,
        suffix: Optional[str] = ".binary",
        length: int,
        padding: int = 256,
    ) -> None:
        loader = ByteSeqLoader(length, padding)
        super().__init__(root, loader, labels, cat=cat, suffix=suffix)
        self.length = length
        self.padding = padding

    def _args(self) -> Tuple[str, ...]:
        return super()._args() + ("length", "padding")


class LabeledImageDataset(LabeledDataset):
    def __init__(
        self,
        root: FilePath,
        labels: pd.DataFrame,
        *,
        cat: bool = True,
        suffix: Optional[str] = ".png",
        loader_mode: Optional[str] = "L",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        loader = PILLoader(loader_mode)
        super().__init__(root, loader, labels, cat=cat, suffix=suffix)
        self.loader_mode = loader_mode
        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index: int) -> Tuple[Any, int]:
        sample, target = super().__getitem__(index)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def _args(self) -> Tuple[str, ...]:
        return super()._args() + ("loader_mode",)


class UnlabeledDataset(Dataset):
    def __init__(
        self,
        root: FilePath,
        loader: Callable[[FilePath], Any],
        *,
        suffix: Optional[str] = None,
    ) -> None:
        self.root = Path(root)
        self.loader = loader
        self.suffix = suffix

        self._sample_paths = sorted(entry for entry in self.root.iterdir())

    def __getitem__(self, index: int) -> Tuple[Any, str]:
        sample_path = self._sample_paths[index]

        if self.suffix is not None:
            sample_name = sample_path.with_suffix(self.suffix).name
        else:
            sample_name = sample_path.name

        sample = self.loader(sample_path)
        return sample, sample_name

    def __len__(self) -> int:
        return len(self._sample_paths)

    def __repr__(self) -> str:
        args = ", ".join(f"{a}={getattr(self, a)!r}" for a in self._args())
        s = f"{type(self).__name__}({args}):\n"
        s += f"    Root directory: {self.root}\n"
        s += f"    Number of samples: {len(self)}\n"
        return s

    def _args(self) -> Tuple[str, ...]:
        return ("suffix",)


class UnlabeledByteSeqDataset(UnlabeledDataset):
    def __init__(
        self,
        root: FilePath,
        *,
        suffix: Optional[str] = "",
        length: int,
        padding: int = 256,
    ) -> None:
        loader = ByteSeqLoader(length, padding)
        super().__init__(root, loader, suffix=suffix)
        self.length = length
        self.padding = padding

    def _args(self) -> Tuple[str, ...]:
        return super()._args() + ("length", "padding")


class UnlabeledImageDataset(UnlabeledDataset):
    def __init__(
        self,
        root: FilePath,
        *,
        suffix: Optional[str] = "",
        loader_mode: Optional[str] = "L",
        transform: Optional[Callable] = None,
    ) -> None:
        loader = PILLoader(loader_mode)
        super().__init__(root, loader, suffix=suffix)
        self.loader_mode = loader_mode
        self.transform = transform

    def __getitem__(self, index: int) -> Tuple[Any, str]:
        sample, sample_name = super().__getitem__(index)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample, sample_name

    def _args(self) -> Tuple[str, ...]:
        return super()._args() + ("loader_mode",)



import functools
import math
from typing import Any, Iterable, Optional, Tuple, cast

import numpy as np
from numpy.typing import NDArray

from ..parallel import reduce_parallel
from ..typing import FilePath

__all__ = [
    "get_byte_seq",
    "extract_byte_seq",
    "extract_byte_seq_parallel",
    "get_byte_hist",
    "extract_byte_hist",
    "extract_byte_hist_parallel",
    "get_byte_entropy_hist",
    "extract_byte_entropy_hist",
    "extract_byte_entropy_hist_parallel",
]


def get_byte_seq(buffer: bytes, *, length: int, padding_value: int = 256) -> NDArray:
    """Get byte sequence."""

    # Read as uint8 but convert to int64 for padding.
    binary = np.frombuffer(buffer[:length], dtype=np.uint8)
    binary = binary.astype(np.int64)
    padding_length = length - len(binary)
    if padding_length > 0:
        padding = np.full(padding_length, padding_value, dtype=np.int64)
        binary = np.concatenate([binary, padding])
    return binary


def extract_byte_seq(binary_file: FilePath, *, length: int, padding_value: int = 256) -> NDArray:
    """Extract byte sequence."""

    with open(binary_file, "rb") as f:
        buffer = f.read(length)
    return get_byte_seq(buffer, length=length, padding_value=padding_value)


def extract_byte_seq_parallel(
    binary_files: Iterable[FilePath],
    *,
    length: int,
    padding_value: int = 256,
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> NDArray:
    """Extract byte sequence in parallel."""

    function = functools.partial(extract_byte_seq, length=length, padding_value=padding_value)
    return reduce_parallel(function, binary_files, n_jobs=n_jobs, **kwargs)


def get_byte_hist(buffer: bytes) -> NDArray:
    """Get byte histogram."""

    binary = np.frombuffer(buffer, dtype=np.uint8)
    return np.bincount(binary, minlength=256)


def extract_byte_hist(binary_file: FilePath) -> NDArray:
    """Extract byte histogram."""

    with open(binary_file, "rb") as f:
        buffer = f.read()
    return get_byte_hist(buffer)


def extract_byte_hist_parallel(
    binary_files: Iterable[FilePath], *, n_jobs: Optional[int] = None, **kwargs: Any
) -> NDArray:
    """Extract byte histogram in parallel."""

    return reduce_parallel(extract_byte_hist, binary_files, n_jobs=n_jobs, **kwargs)


def _get_entropy_and_byte_hist(window: NDArray) -> Tuple[float, NDArray]:
    byte_hist = np.bincount(window, minlength=256)
    prob = byte_hist / len(window)
    mask = prob > 0
    # Create zeros out array to prevent `np.log2()` from creating uninitialized out array.
    out = np.zeros_like(prob)
    entropy = -np.sum(prob * np.log2(prob, out=out, where=mask))
    return float(entropy), byte_hist


def get_byte_entropy_hist(buffer: bytes, *, window_size: int = 1024, step_size: int = 256) -> NDArray:
    """Get byte entropy histogram.

    References:
        - J. Saxe and K. Berlin. 2015.
          Deep neural network based malware detection using two dimensional binary program features.
          https://doi.org/10.1109/MALWARE.2015.7413680
    """

    binary = np.frombuffer(buffer, dtype=np.uint8)
    matrix = np.zeros((16, 256), dtype=np.int64)
    # https://numpy.org/doc/stable/reference/generated/numpy.lib.stride_tricks.sliding_window_view.html
    for window in np.lib.stride_tricks.sliding_window_view(binary, window_size)[::step_size]:
        entropy, byte_hist = _get_entropy_and_byte_hist(window)
        index = min(math.floor(entropy * 2), 15)
        matrix[index] += byte_hist
    result = cast(NDArray, matrix.reshape((16, 16, 16)).sum(-1))
    return result.reshape(256)


def extract_byte_entropy_hist(binary_file: FilePath, *, window_size: int = 1024, step_size: int = 256) -> NDArray:
    """Extract byte entropy histogram."""

    with open(binary_file, "rb") as f:
        buffer = f.read()
    return get_byte_entropy_hist(buffer, window_size=window_size, step_size=step_size)


def extract_byte_entropy_hist_parallel(
    binary_files: Iterable[FilePath],
    *,
    window_size: int = 1024,
    step_size: int = 256,
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> NDArray:
    """Extract byte entropy histogram in parallel."""

    function = functools.partial(extract_byte_entropy_hist, window_size=window_size, step_size=step_size)
    return reduce_parallel(function, binary_files, n_jobs=n_jobs, **kwargs)



from typing import Iterable, Optional, Tuple

from numpy.typing import NDArray

from ..features.binary import (
    extract_byte_entropy_hist,
    extract_byte_entropy_hist_parallel,
    extract_byte_hist,
    extract_byte_hist_parallel,
    extract_byte_seq,
    extract_byte_seq_parallel,
    get_byte_entropy_hist,
    get_byte_hist,
    get_byte_seq,
)
from ..typing import FilePath
from ._loader import Loader

__all__ = ["ByteSeqLoader", "ByteHistLoader", "ByteEntropyHistLoader"]


class ByteSeqLoader(Loader):
    def __init__(self, length: int, padding_value: int = 256) -> None:
        self.length = length
        self.padding_value = padding_value

    def __call__(self, path: FilePath) -> NDArray:
        return extract_byte_seq(path, length=self.length, padding_value=self.padding_value)

    def parallel(self, paths: Iterable[FilePath], *, n_jobs: Optional[int] = None) -> NDArray:
        return extract_byte_seq_parallel(paths, length=self.length, padding_value=self.padding_value, n_jobs=n_jobs)

    def from_buffer(self, buffer: bytes) -> NDArray:
        return get_byte_seq(buffer, length=self.length, padding_value=self.padding_value)

    def _get_args(self) -> Tuple[str, ...]:
        return ("length", "padding_value")


class ByteHistLoader(Loader):
    def __call__(self, path: FilePath) -> NDArray:
        return extract_byte_hist(path)

    def parallel(self, paths: Iterable[FilePath], *, n_jobs: Optional[int] = None) -> NDArray:
        return extract_byte_hist_parallel(paths, n_jobs=n_jobs)

    def from_buffer(self, buffer: bytes) -> NDArray:
        return get_byte_hist(buffer)

    def _get_args(self) -> Tuple[str, ...]:
        return ()


class ByteEntropyHistLoader(Loader):
    def __init__(self, window_size: int = 1024, step_size: int = 256) -> None:
        self.window_size = window_size
        self.step_size = step_size

    def __call__(self, path: FilePath) -> NDArray:
        return extract_byte_entropy_hist(path, window_size=self.window_size, step_size=self.step_size)

    def parallel(self, paths: Iterable[FilePath], *, n_jobs: Optional[int] = None) -> NDArray:
        return extract_byte_entropy_hist_parallel(
            paths, window_size=self.window_size, step_size=self.step_size, n_jobs=n_jobs
        )

    def from_buffer(self, buffer: bytes) -> NDArray:
        return get_byte_entropy_hist(buffer, window_size=self.window_size, step_size=self.step_size)

    def _get_args(self) -> Tuple[str, ...]:
        return ("window_size", "step_size")



import functools
from typing import Any, Callable, Dict, Iterable, Optional, Union

from PIL import Image

from ..parallel import execute_parallel
from ..typing import FilePath

__all__ = [
    "get_image",
    "convert_binary_to_image",
    "convert_binary_to_image_parallel",
    "resize_image",
    "resize_image_parallel",
]

_registry: Dict[str, Callable[[int], int]] = {}


def register_width_function(name: str) -> Callable[[Callable[[int], int]], Callable[[int], int]]:
    """Register width function."""

    def decorator(func: Callable[[int], int]) -> Callable[[int], int]:
        _registry[name] = func
        return func

    return decorator


@register_width_function("nataraj")
def width_function_nataraj(filesize: int) -> int:
    """Set width depending on filesize.

    References:
        - Nataraj et al. 2011. Malware images: visualization and automatic classification.
          https://doi.org/10.1145/2016904.2016908
    """

    k = filesize // 1024
    if k < 10:
        width = 32
    elif k < 30:
        width = 64
    elif k < 60:
        width = 128
    elif k < 100:
        width = 256
    elif k < 200:
        width = 384
    elif k < 500:
        width = 512
    elif k < 1000:
        width = 768
    else:
        width = 1024
    return width


def _check_width(filesize: int, width: Union[int, str, Callable[[int], int]]) -> int:
    if isinstance(width, str):
        try:
            width = _registry[width](filesize)
        except KeyError:
            keys = "{" + ", ".join(_registry.keys()) + "}"
            raise ValueError(f"unknown width, expected {keys}")
    elif callable(width):
        width = width(filesize)
    return width


def get_image(
    buffer: bytes, *, width: Union[int, str, Callable[[int], int]], drop: bool = False, padding: bytes = b"\x00"
) -> Image.Image:
    """Get image."""

    nbytes = len(buffer)
    width = _check_width(nbytes, width)
    height, rem = divmod(nbytes, width)
    if rem != 0:
        # The image height must be at least 1.
        if drop and height > 0:
            nbytes = width * height
            buffer = buffer[:nbytes]
        else:
            height += 1
            nbytes = width * height
            buffer = buffer.ljust(nbytes, padding)
    return Image.frombuffer("L", (width, height), buffer, "raw", "L", 0, 1)


def convert_binary_to_image(
    binary_file: FilePath,
    image_file: FilePath,
    *,
    width: Union[int, str, Callable[[int], int]],
    drop: bool = False,
    padding: bytes = b"\x00",
) -> None:
    """Convert binary file to image file."""

    with open(binary_file, "rb") as f:
        buffer = f.read()
    image = get_image(buffer, width=width, drop=drop, padding=padding)
    image.save(image_file)


def convert_binary_to_image_parallel(
    binary_files: Iterable[FilePath],
    image_files: Iterable[FilePath],
    *,
    width: Union[int, str, Callable[[int], int]],
    drop: bool = False,
    padding: bytes = b"\x00",
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> None:
    """Convert binary file to image file in parallel."""

    function = functools.partial(convert_binary_to_image, width=width, drop=drop, padding=padding)
    execute_parallel(function, binary_files, image_files, n_jobs=n_jobs, **kwargs)


def resize_image(src: FilePath, dst: FilePath, *, width: int, height: int) -> None:
    """Resize image file."""

    image = Image.open(src)
    image = image.resize((width, height), Image.BILINEAR)
    image.save(dst)


def resize_image_parallel(
    srcs: Iterable[FilePath],
    dsts: Iterable[FilePath],
    *,
    width: int,
    height: int,
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> None:
    """Resize image file in parallel."""

    function = functools.partial(resize_image, width=width, height=height)
    execute_parallel(function, srcs, dsts, n_jobs=n_jobs, **kwargs)



import functools
import shutil
from pathlib import Path
from typing import Any, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd

from .parallel import execute_parallel
from .typing import FilePath

__all__ = [
    "categorize_folders",
    "split_labels",
    "build_srcs_dsts",
    "convert_bytes_to_binary",
    "convert_bytes_to_binary_parallel",
]


def categorize_folders(root: FilePath, labels: pd.DataFrame, *, suffix: Optional[str] = None) -> None:
    """Categorize samples and move them into class name folders."""

    root = Path(root)
    sample_names = [str(item) for item in labels.iloc[:, 0]]
    target_names = [str(item) for item in labels.iloc[:, 1]]

    for klass in set(target_names):
        cat_dir = root / klass
        if cat_dir.is_dir():
            return
        cat_dir.mkdir(parents=True, exist_ok=True)

    for sample_name, target_name in zip(sample_names, target_names):
        src = root / sample_name
        if suffix is not None:
            src = src.with_suffix(suffix)
        shutil.move(src, root / target_name)


def split_labels(
    labels: pd.DataFrame,
    *,
    test_size: Optional[float] = None,
    train_size: Optional[float] = None,
    random_state: Optional[int] = None,
    shuffle: bool = True,
    stratified: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Split labels into two parts."""

    from sklearn.model_selection import train_test_split

    stratify = None
    if shuffle and stratified:
        stratify = labels.iloc[:, 1].to_numpy()

    indices = np.arange(len(labels))
    idx1, idx2 = train_test_split(
        indices,
        test_size=test_size,
        train_size=train_size,
        random_state=random_state,
        shuffle=shuffle,
        stratify=stratify,
    )
    return labels.iloc[idx1], labels.iloc[idx2]


def _build_srcs_dsts(
    src_dir: Path, dst_dir: Path, *, skip_exist: bool = True, suffix: Optional[str] = None
) -> Tuple[List[Path], List[Path]]:
    srcs = []
    dsts = []
    dst_dir.mkdir(parents=True, exist_ok=True)
    for src in src_dir.iterdir():
        if suffix is None:
            name = src.name
        else:
            name = src.with_suffix(suffix).name
        dst = dst_dir / name
        if skip_exist and dst.exists():
            continue
        srcs.append(src)
        dsts.append(dst)
    return srcs, dsts


def _build_srcs_dsts_cat(
    src_dir: Path, dst_dir: Path, *, skip_exist: bool = True, suffix: Optional[str] = None
) -> Tuple[List[Path], List[Path]]:
    srcs = []
    dsts = []
    dst_dir.mkdir(parents=True, exist_ok=True)
    for src_cat_dir in src_dir.iterdir():
        dst_cat_dir = dst_dir / src_cat_dir.name
        dst_cat_dir.mkdir(parents=True, exist_ok=True)
        for src in src_cat_dir.iterdir():
            if suffix is None:
                name = src.name
            else:
                name = src.with_suffix(suffix).name
            dst = dst_cat_dir / name
            if skip_exist and dst.exists():
                continue
            srcs.append(src)
            dsts.append(dst)
    return srcs, dsts


def build_srcs_dsts(
    src_dir: FilePath, dst_dir: FilePath, *, cat: bool = True, skip_exist: bool = True, suffix: Optional[str] = None
) -> Tuple[List[Path], List[Path]]:
    """Build source paths and destination paths."""

    src_dir = Path(src_dir)
    dst_dir = Path(dst_dir)
    if cat:
        return _build_srcs_dsts_cat(src_dir, dst_dir, suffix=suffix, skip_exist=skip_exist)
    else:
        return _build_srcs_dsts(src_dir, dst_dir, suffix=suffix, skip_exist=skip_exist)


def convert_bytes_to_binary(bytes_file: FilePath, binary_file: FilePath, *, qq: str = "00") -> None:
    """Convert bytes file to binary file."""

    with open(bytes_file, encoding="ascii") as src, open(binary_file, "wb") as dst:
        for line in src:
            i = line.find(" ")
            if i < 0:
                raise ValueError(f"invalid bytes file {bytes_file!r}")
            data = line[i + 1 :].replace("??", qq)
            dst.write(bytes.fromhex(data))


def convert_bytes_to_binary_parallel(
    bytes_files: Iterable[FilePath],
    binary_files: Iterable[FilePath],
    *,
    qq: str = "00",
    n_jobs: Optional[int] = None,
    **kwargs: Any,
) -> None:
    """Convert bytes file to binary file in parallel."""

    function = functools.partial(convert_bytes_to_binary, qq=qq)
    execute_parallel(function, bytes_files, binary_files, n_jobs=n_jobs, **kwargs)



from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..typing import FilePath
from ._loader import Loader


class ClassifiedDataset:
    def __init__(
        self,
        root: FilePath,
        label: pd.DataFrame,
        loader: Loader,
        *,
        num_classes: Optional[int] = None,
    ) -> None:
        root = Path(root)

        if not root.is_dir():
            raise ValueError(f"{root} is not a directory")

        self.root = root
        self.label = label
        self.loader = loader

        self._sample_names = [str(item) for item in label.iloc[:, 0]]
        self._target_names = [str(item) for item in label.iloc[:, 1]]

        self._index_to_class = sorted(p.name for p in root.iterdir() if p.is_dir())
        self._class_to_index = {c: i for i, c in enumerate(self._index_to_class)}

        if num_classes is not None and len(self._class_to_index) != num_classes:
            raise RuntimeError("num_classes check failure, the root directory may be dirty")

    @property
    def sample_names(self) -> List[str]:
        return self._sample_names

    @property
    def target_names(self) -> List[str]:
        return self._target_names

    @property
    def index_to_class(self) -> List[str]:
        return self._index_to_class

    @property
    def class_to_index(self) -> Dict[str, int]:
        return self._class_to_index

    def __len__(self) -> int:
        return len(self.sample_names)

    @property
    def num_classes(self) -> int:
        return len(self.class_to_index)

    def __iter__(self) -> Iterator[Tuple[NDArray, int]]:
        for index in range(len(self)):
            yield self[index]

    def __getitem__(self, index: int) -> Tuple[NDArray, int]:
        sample_name = self.sample_names[index]
        target_name = self.target_names[index]
        sample_path = self.root / target_name / sample_name
        sample = self.loader(sample_path)
        target = self.class_to_index[target_name]
        return sample, target

    def get_sample_path(self, index: int) -> Path:
        sample_name = self.sample_names[index]
        target_name = self.target_names[index]
        sample_path = self.root / target_name / sample_name
        return sample_path

    def get_sample_path_alter(self, index: int, root: FilePath) -> Path:
        sample_name = self.sample_names[index]
        target_name = self.target_names[index]
        sample_dir = Path(root) / target_name
        sample_dir.mkdir(parents=True, exist_ok=True)
        sample_path = sample_dir / sample_name
        return sample_path

    def get_sample(self, index: int) -> NDArray:
        sample_path = self.get_sample_path(index)
        sample = self.loader(sample_path)
        return sample

    def get_target(self, index: int) -> int:
        target_name = self.target_names[index]
        target = self.class_to_index[target_name]
        return target

    def load_X_y(self) -> Tuple[NDArray, NDArray]:
        X = []
        y = []
        for sample, target in self:
            X.append(sample)
            y.append(target)
        return np.stack(X), np.stack(y)

    def load_X_y_parallel(self, *, n_jobs: Optional[int] = None) -> Tuple[NDArray, NDArray]:
        sample_paths = []
        y = []
        for index in range(len(self)):
            sample_name = self.sample_names[index]
            target_name = self.target_names[index]
            sample_path = self.root / target_name / sample_name
            target = self.class_to_index[target_name]
            sample_paths.append(sample_path)
            y.append(target)
        X = self.loader.parallel(sample_paths, n_jobs=n_jobs)
        return X, np.stack(y)

    def __repr__(self) -> str:
        s = f"{type(self).__name__}:\n"
        s += f"    Root: {self.root}\n"
        s += f"    Loader: {self.loader}\n"
        s += f"    Number of samples: {len(self)}\n"
        s += f"    Number of classes: {self.num_classes}\n"
        return s



from . import models
from .checkpoint import Checkpoint
from .dataset import (
    LabeledByteSeqDataset,
    LabeledDataset,
    LabeledImageDataset,
    UnlabeledByteSeqDataset,
    UnlabeledDataset,
    UnlabeledImageDataset,
)

__all__ = [
    "models",
    "Checkpoint",
    "LabeledDataset",
    "LabeledByteSeqDataset",
    "LabeledImageDataset",
    "UnlabeledDataset",
    "UnlabeledByteSeqDataset",
    "UnlabeledImageDataset",
]



from typing import Optional

import torch
import torch.nn as nn
from torch import Tensor


class MalConv(nn.Module):
    """The MalConv model.

    References:
        - Edward Raff et al. 2018. Malware Detection by Eating a Whole EXE.
          https://arxiv.org/abs/1710.09435
    """

    def __init__(
        self,
        num_classes: int = 2,
        *,
        num_embeddings: int = 257,
        embedding_dim: int = 8,
        channels: int = 128,
        kernel_size: int = 512,
        stride: int = 512,
        padding_idx: Optional[int] = 256,
    ) -> None:
        super().__init__()

        self.num_classes = num_classes

        # By default, num_embeddings (257) = byte (0-255) + padding (256).
        self.embedding = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)

        self.conv1 = nn.Conv1d(embedding_dim, channels, kernel_size=kernel_size, stride=stride, bias=True)
        self.conv2 = nn.Conv1d(embedding_dim, channels, kernel_size=kernel_size, stride=stride, bias=True)

        self.maxpool = nn.AdaptiveMaxPool1d(1)

        self.fc = nn.Sequential(
            nn.Flatten(),
            nn.Linear(channels, channels),
            nn.ReLU(inplace=True),
            nn.Linear(channels, num_classes),
        )

    def _embed(self, x: Tensor) -> Tensor:
        # Perform embedding.
        x = self.embedding(x)

        # Treat embedding dimension as channel.
        x = x.permute(0, 2, 1)
        return x

    def _forward_embedded(self, x: Tensor) -> Tensor:
        # Perform gated convolution.
        x = self.conv1(x) * torch.sigmoid(self.conv2(x))

        # Perform global max pooling.
        x = self.maxpool(x)

        x = self.fc(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        x = self._embed(x)
        x = self._forward_embedded(x)
        return x



from typing import Iterable, List, Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from ..parallel import reduce_parallel
from ..typing import FilePath


class Loader:
    def __call__(self, path: FilePath) -> NDArray:
        raise NotImplementedError

    def parallel(self, paths: Iterable[FilePath], *, n_jobs: Optional[int] = None) -> NDArray:
        raise NotImplementedError

    def from_buffer(self, buffer: bytes) -> NDArray:
        raise NotImplementedError

    def __repr__(self) -> str:
        # Generate pretty formatted arguments.
        args = self._get_args()
        args_str = ", ".join(f"{a}={getattr(self, a)!r}" for a in args)
        return f"{type(self).__name__}({args_str})"

    def _get_args(self) -> Tuple[str, ...]:
        raise NotImplementedError


class MultiLoader(Loader):
    def __init__(self, loaders: List[Loader]) -> None:
        self.loaders = loaders

    def __call__(self, path: FilePath) -> NDArray:
        with open(path, "rb") as f:
            buffer = f.read()
        return self.from_buffer(buffer)

    def parallel(self, paths: Iterable[FilePath], *, n_jobs: Optional[int] = None) -> NDArray:
        return reduce_parallel(self, paths, n_jobs=n_jobs)

    def from_buffer(self, buffer: bytes) -> NDArray:
        return np.concatenate([loader.from_buffer(buffer) for loader in self.loaders])

    def _get_args(self) -> Tuple[str, ...]:
        args: List[str] = []
        for loader in self.loaders:
            args.extend(loader._get_args())
        return tuple(args)



import operator
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, List, Mapping, Optional, Union

import torch
import torch.nn as nn

from ..typing import FilePath

__all__ = ["Checkpoint"]


@dataclass(order=True)
class Item:
    priority: Union[int, float]
    filename: str = field(compare=False)


class PriorityQueue:
    def __init__(self, *, capacity: int = 1, reverse: bool = False) -> None:
        if capacity < 1:
            raise ValueError(f"require capacity >= 1, got {capacity!r}")

        self._queue: List[Item] = []
        self._capacity = capacity
        self._reverse = reverse

        if reverse:
            self._cmp = operator.gt
        else:
            self._cmp = operator.lt

    def __len__(self) -> int:
        return len(self._queue)

    @property
    def capacity(self) -> int:
        return self._capacity

    def check_priority(self, priority: Union[int, float]) -> bool:
        if len(self) < self.capacity:
            return True
        return bool(self._cmp(priority, self._queue[-1].priority))

    def update(self, priority: Union[int, float], filename: str) -> Optional[Item]:
        last = None
        if len(self) == self.capacity:
            last = self.pop()
        self.push(priority, filename)
        return last

    def push(self, priority: Union[int, float], filename: str) -> None:
        if len(self) == self.capacity:
            raise ValueError("queue is full")
        item = Item(priority, filename)
        self._queue.append(item)
        self._queue.sort(reverse=self._reverse)

    def pop(self) -> Item:
        if not self:
            raise ValueError("queue is empty")
        return self._queue.pop()


class Checkpoint:
    def __init__(self, root: FilePath, filename_prefix: str, max_save: int = 1, reverse: bool = False) -> None:
        self.root = Path(root)
        self.filename_prefix = filename_prefix
        self.max_save = max_save
        self.queue = PriorityQueue(capacity=max_save, reverse=reverse)

        self.root.mkdir(parents=True, exist_ok=True)

    def save(
        self,
        check_dict: Mapping[str, Any],
        metric_name: str,
        metric_value: Union[int, float],
        global_step: Optional[int] = None,
    ) -> None:
        if not self.queue.check_priority(metric_value):
            return

        filename = self.filename_prefix
        if global_step is not None:
            filename += f"_{global_step}"
        if isinstance(metric_value, int):
            filename += f"_{metric_name}={metric_value}.pt"
        else:
            filename += f"_{metric_name}={metric_value:.4f}.pt"

        state_dict = {}
        for key, obj in check_dict.items():
            if isinstance(obj, (nn.DataParallel, nn.parallel.DistributedDataParallel)):
                obj = obj.module
            state_dict[key] = obj.state_dict()
        torch.save(state_dict, self.root / filename)

        old = self.queue.update(metric_value, filename)
        if old is not None:
            oldpath = self.root / old.filename
            oldpath.unlink(missing_ok=True)

    @staticmethod
    def load(check_dict: Mapping[str, Any], filepath: FilePath, map_location: Any = None) -> None:
        state_dict = torch.load(filepath, map_location=map_location)
        for key, obj in check_dict.items():
            if key not in state_dict:
                raise ValueError(f"{key} does not exist")
            if isinstance(obj, (nn.DataParallel, nn.parallel.DistributedDataParallel)):
                obj = obj.module
            obj.load_state_dict(state_dict[key])
